@@ -1,38 +1,44 @@
 import asyncio
-
+from typing import List, Dict
 
 from msldap import logger
 from msldap.commons.common import MSLDAPClientStatus
+from .commons.target import MSLDAPTarget
 from msldap.protocol.messages import LDAPMessage, BindRequest, \
 	protocolOp, AuthenticationChoice, SaslCredentials, \
 	SearchRequest, AttributeDescription, Filter, Filters, \
 	Controls, Control, SearchControlValue, AddRequest, \
-	ModifyRequest, DelRequest
+	ModifyRequest, DelRequest, ExtendedRequest, ExtendedResponse
 
 from msldap.protocol.utils import calcualte_length
 from msldap.protocol.typeconversion import convert_result, convert_attributes, encode_attributes, encode_changes
 from msldap.protocol.query import escape_filter_chars, query_syntax_converter
-from msldap.commons.authbuilder import AuthenticatorBuilder
-from msldap.commons.credential import MSLDAP_GSS_METHODS
-from msldap.network.selector import MSLDAPNetworkSelector
-from msldap.commons.credential import LDAPAuthProtocol
-from msldap.commons.target import LDAPProtocol
-from asn1crypto.x509 import Certificate
+from msldap.commons.authbuilder import get_auth_context
+from msldap.network.packetizer import LDAPPacketizer
+from asysocks.unicomm.common.target import UniProto
+from msldap.commons.exceptions import LDAPBindException, LDAPAddException, LDAPModifyException, LDAPDeleteException, LDAPSearchException
 from hashlib import sha256
-from minikerberos.gssapi.channelbindings import ChannelBindingsStruct
+from asysocks.unicomm.client import UniClient
+from asyauth.common.constants import asyauthProtocol
+from asyauth.common.credentials import UniCredential
+from asyauth.common.winapi.constants import ISC_REQ
 
 class MSLDAPClientConnection:
-	def __init__(self, target, creds):
-		if target is None:
-			raise Exception('Target cant be none!')
+	def __init__(self, target:MSLDAPTarget, credential:UniCredential, auth=None):
 		self.target = target
-		self.creds = creds
-		self.auth = AuthenticatorBuilder(self.creds, self.target).build()
+		self.credential = credential
+		if auth is not None:
+			self.auth = auth
+		else:
+			self.auth = get_auth_context(self.credential)
+
 		self.connected = False
 		self.bind_ok = False
+		self.is_anon = False
 		self.__sign_messages = False
 		self.__encrypt_messages = False
 		self.network = None
+		self.connection_closed_evt = None
 
 		self.handle_incoming_task = None
 		self.status = MSLDAPClientStatus.RUNNING
@@ -44,14 +50,19 @@ class MSLDAPClientConnection:
 		self.encryption_sequence_counter = 0 # this will be set by the inderlying auth algo
 		self.cb_data = None #for channel binding
 
+		if self.credential.protocol == asyauthProtocol.NONE:
+			self.is_anon = True
+	
+	async def __aenter__(self):
+		return self
+		
+	async def __aexit__(self, exc_type, exc, traceback):
+		await asyncio.wait_for(self.disconnect(), timeout = 1)
+
 	async def __handle_incoming(self):
 		try:
-			while True:
-				message_data, err = await self.network.in_queue.get()
-				if err is not None:
-					logger.debug('Client terminating bc __handle_incoming got an error!')
-					raise err
-				
+			
+			async for message_data in self.network.read():				
 				#print('Incoming message data: %s' % message_data)
 				if self.bind_ok is True:
 					if self.__encrypt_messages is True:
@@ -116,10 +127,14 @@ class MSLDAPClientConnection:
 				self.message_table[msgid] = [e]
 				self.message_table_notify[msgid].set()
 		
-		self.status = MSLDAPClientStatus.STOPPED
+		finally:
+			if self.connection_closed_evt is not None:
+				self.connection_closed_evt.set()
+			self.status = MSLDAPClientStatus.STOPPED
 
 
-	async def send_message(self, message):
+
+	async def send_message(self, message:Dict[str, object]):
 		curr_msg_id = self.message_id
 		self.message_id += 1
 
@@ -139,11 +154,11 @@ class MSLDAPClientConnection:
 				self.encryption_sequence_counter += 1
 		
 		self.message_table_notify[curr_msg_id] = asyncio.Event()
-		await self.network.out_queue.put(message_data)
+		await self.network.write(message_data)
 
 		return curr_msg_id
 
-	async def recv_message(self, message_id):
+	async def recv_message(self, message_id:int):
 		if message_id not in self.message_table_notify:
 			logger.debug('Requested message id %s which is not in the message notify table!' % message_id)
 			return None
@@ -169,22 +184,18 @@ class MSLDAPClientConnection:
 		"""
 		try:
 			logger.debug('Connecting!')
-			self.network = await MSLDAPNetworkSelector.select(self.target)
-			res, err = await self.network.run()
-			if res is False:
-				return False, err
+			self.connection_closed_evt = asyncio.Event()
+			packetizer = LDAPPacketizer()
+			client = UniClient(self.target, packetizer)
+			self.network = await client.connect()
 			
 			# now processing channel binding options
-			if self.target.proto == LDAPProtocol.SSL:
+			if self.target.protocol == UniProto.CLIENT_SSL_TCP:
 				certdata = self.network.get_peer_certificate()
-				#cert = Certificate.load(certdata).native
-				#print(cert)
-				cb_struct = ChannelBindingsStruct()
-				cb_struct.application_data = b'tls-server-end-point:' + sha256(certdata).digest()
-
-				self.cb_data = cb_struct.to_bytes()
+				self.cb_data = b'tls-server-end-point:' + sha256(certdata).digest()
 
 			self.handle_incoming_task = asyncio.create_task(self.__handle_incoming())
+			self.status = MSLDAPClientStatus.CONNECTED
 			logger.debug('Connection succsessful!')
 			return True, None
 		except Exception as e:
@@ -200,10 +211,15 @@ class MSLDAPClientConnection:
 
 		logger.debug('Disconnecting!')
 		self.bind_ok = False
+		if self.connection_closed_evt is not None:
+			self.connection_closed_evt.set()
+		
+		if self.network is not None:
+			await self.network.close()
+			await asyncio.sleep(0)
+		
 		if self.handle_incoming_task is not None:
 			self.handle_incoming_task.cancel()
-		if self.network is not None:
-			await self.network.terminate()
 
 
 	def __bind_success(self):
@@ -213,11 +229,12 @@ class MSLDAPClientConnection:
 		"""
 		logger.debug('BIND Success!')
 		self.bind_ok = True
-		if self.creds.auth_method in MSLDAP_GSS_METHODS or self.creds.auth_method == LDAPAuthProtocol.SICILY:
+		self.status = MSLDAPClientStatus.RUNNING
+		if self.credential.protocol in [asyauthProtocol.NTLM, asyauthProtocol.KERBEROS, asyauthProtocol.SICILY]:
 			self.__sign_messages = self.auth.signing_needed()
 			self.__encrypt_messages = self.auth.encryption_needed()
 			if self.__encrypt_messages or self.__sign_messages:
-				self.network.is_plain_msg = False
+				self.network.packetizer.is_plain_msg = False
 
 	async def bind(self):
 		"""
@@ -229,9 +246,11 @@ class MSLDAPClientConnection:
 		"""
 		logger.debug('BIND in progress...')
 		try:
-			if self.creds.auth_method == LDAPAuthProtocol.SICILY:
-				
-				data, to_continue, err = await self.auth.authenticate(None)
+			if self.credential.protocol == asyauthProtocol.SICILY:
+				flags = ISC_REQ.CONNECTION|ISC_REQ.CONFIDENTIALITY|ISC_REQ.INTEGRITY
+				if self.target.protocol == UniProto.CLIENT_SSL_TCP:
+					flags = ISC_REQ.CONNECTION
+				data, to_continue, err = await self.auth.authenticate(None, spn=self.target.to_target_string(), flags=flags, cb_data = self.cb_data)
 				if err is not None:
 					return None, err
 
@@ -255,11 +274,10 @@ class MSLDAPClientConnection:
 					return False, res
 				res = res.native
 				if res['protocolOp']['resultCode'] != 'success':
-					return False, Exception(
-						'BIND failed! Result code: "%s" Reason: "%s"' % (
+					return False, LDAPBindException(
 							res['protocolOp']['resultCode'], 
 							res['protocolOp']['diagnosticMessage']
-						))
+						)
 				
 				auth = {
 					'sicily_nego' : data
@@ -281,13 +299,12 @@ class MSLDAPClientConnection:
 					return False, res
 				res = res.native
 				if res['protocolOp']['resultCode'] != 'success':
-					return False, Exception(
-						'BIND failed! Result code: "%s" Reason: "%s"' % (
+					return False, LDAPBindException(
 							res['protocolOp']['resultCode'], 
 							res['protocolOp']['diagnosticMessage']
-						))
+						)
 
-				data, to_continue, err = await self.auth.authenticate(res['protocolOp']['matchedDN'])
+				data, to_continue, err = await self.auth.authenticate(res['protocolOp']['matchedDN'], spn=self.target.to_target_string(), cb_data = self.cb_data)
 				if err is not None:
 					return None, err
 
@@ -311,25 +328,26 @@ class MSLDAPClientConnection:
 					return False, res
 				res = res.native
 				if res['protocolOp']['resultCode'] != 'success':
-					return False, Exception(
-						'BIND failed! Result code: "%s" Reason: "%s"' % (
+					return False, LDAPBindException(
 							res['protocolOp']['resultCode'], 
 							res['protocolOp']['diagnosticMessage']
-						))
+						)
 				
 
 				self.__bind_success()
 				return True, None
 
-			elif self.creds.auth_method == LDAPAuthProtocol.SIMPLE:
+			elif self.credential.protocol == asyauthProtocol.SIMPLE:
 				pw = b''
-				if self.auth.password != None:
-					pw = self.auth.password.encode()
+				if self.auth.secret != None:
+					pw = self.auth.secret.encode()
 
 				user = b''
 				if self.auth.username != None:
 					user = self.auth.username.encode()
-
+				if user == b'':
+					self.is_anon = True
+				
 				auth = {
 					'simple' : pw
 				}
@@ -354,17 +372,20 @@ class MSLDAPClientConnection:
 					return True, None
 				
 				else:
-					return False, Exception(
-						'BIND failed! Result code: "%s" Reason: "%s"' % (
+					return False, LDAPBindException(
 							res['protocolOp']['resultCode'], 
 							res['protocolOp']['diagnosticMessage']
-						))
+						)
 
-			elif self.creds.auth_method in MSLDAP_GSS_METHODS:
+			elif self.credential.protocol in [asyauthProtocol.NTLM, asyauthProtocol.KERBEROS]:
 				challenge = None
 				while True:
 					try:
-						data, to_continue, err = await self.auth.authenticate(challenge, cb_data = self.cb_data)
+						flags = ISC_REQ.CONNECTION|ISC_REQ.CONFIDENTIALITY|ISC_REQ.INTEGRITY
+						if self.target.protocol == UniProto.CLIENT_SSL_TCP:
+							flags = ISC_REQ.CONNECTION
+						
+						data, to_continue, err = await self.auth.authenticate(challenge, cb_data = self.cb_data, spn=self.target.to_target_string(), flags=flags)
 						if err is not None:
 							raise err
 					except Exception as e:
@@ -395,11 +416,12 @@ class MSLDAPClientConnection:
 					res = res.native
 					if res['protocolOp']['resultCode'] == 'success':
 						if 'serverSaslCreds' in res['protocolOp']:
-							data, _, err = await self.auth.authenticate(res['protocolOp']['serverSaslCreds'], cb_data = self.cb_data)
+							data, _, err = await self.auth.authenticate(res['protocolOp']['serverSaslCreds'], cb_data = self.cb_data, spn=self.target.to_target_string(), flags=flags)
 							if err is not None:
 								return False, err
 
-						self.encryption_sequence_counter = self.auth.get_seq_number()
+						if self.auth.encryption_needed() is True or self.auth.signing_needed() is True:
+							self.encryption_sequence_counter = self.auth.get_seq_number()
 						self.__bind_success()
 
 						return True, None
@@ -409,18 +431,89 @@ class MSLDAPClientConnection:
 						continue
 
 					else:
-						return False, Exception(
-							'BIND failed! Result code: "%s" Reason: "%s"' % (
+						return False, LDAPBindException(
 								res['protocolOp']['resultCode'], 
 								res['protocolOp']['diagnosticMessage']
-							))
+							)
+			elif self.credential.protocol == asyauthProtocol.SSL:
+				if self.target.protocol == UniProto.CLIENT_SSL_TCP:
+					self.__bind_success()
+					return True, None
+				elif self.target.protocol == UniProto.CLIENT_TCP:
+					_, err = await self.starttls()
+					if err is not None:
+						return False, err
+					
+					sasl = {
+						'mechanism' : 'EXTERNAL'.encode(),
+					}
+					auth = {
+						'sasl' : SaslCredentials(sasl)
+					}
+
+					bindreq = {
+						'version' : 3,
+						'name': b'',
+						'authentication': AuthenticationChoice(auth), 
+					}
+
+					br = { 'bindRequest' : BindRequest( bindreq	)}
+					msg = { 'protocolOp' : protocolOp(br)}
+
+					msg_id = await self.send_message(msg)
+					res = await self.recv_message(msg_id)
+					if isinstance(res[0], Exception):
+						return False, res[0]
+					resp = res[0].native['protocolOp']
+					if resp['resultCode'] != 'success':
+						#startls refused :(
+						return False, LDAPBindException(
+								resp['resultCode'], 
+								resp['diagnosticMessage']
+							)
+					self.__bind_success()
+					return True, None
+				
+				return False, Exception('Not implemented: %s' % self.credential.protocol.name)
 					
 			else:
-				raise Exception('Not implemented authentication method: %s' % self.creds.auth_method.name)
+				raise Exception('Not implemented authentication method: %s' % self.credential.protocol.name)
+			
 		except Exception as e:
+			await self.disconnect()
 			return False, e
 
-	async def add(self, entry, attributes):
+	async def starttls(self):
+		try:
+			# https://www.ietf.org/rfc/rfc2830.txt
+			ext = {
+				'requestName': b'1.3.6.1.4.1.1466.20037', 
+			}
+			br = { 'extendedReq' : ExtendedRequest(ext)}
+			msg = { 'protocolOp' : protocolOp(br)}
+			msg_id = await self.send_message(msg)
+			res = await self.recv_message(msg_id)
+			if isinstance(res[0], Exception):
+				return False, res[0]
+			resp = res[0].native['protocolOp']
+			if resp['resultCode'] != 'success':
+				#startls refused :(
+				return False, LDAPBindException(
+						resp['resultCode'], 
+						resp['diagnosticMessage']
+					)
+			self.handle_incoming_task.cancel()
+			await asyncio.sleep(0)
+			await self.network.wrap_ssl(self.target.get_ssl_context())
+			self.handle_incoming_task = asyncio.create_task(self.__handle_incoming())
+			self.status = MSLDAPClientStatus.CONNECTED
+			await asyncio.sleep(0)
+			return True, None
+		except Exception as e:
+			print(e)
+			return False, e
+
+	async def add(self, entry:str, attributes:Dict[str, object]):
 		"""
 		Performs the add operation.
 		
@@ -436,6 +529,7 @@ class MSLDAPClientConnection:
 				'entry' : entry.encode(),
 				'attributes' : encode_attributes(attributes)
 			}
+			logger.debug(req)
 			br = { 'addRequest' : AddRequest(req)}
 			msg = { 'protocolOp' : protocolOp(br)}
 			
@@ -449,16 +543,17 @@ class MSLDAPClientConnection:
 				message = message.native
 				if msg_type == 'addResponse':
 					if message['protocolOp']['resultCode'] != 'success':
-						return False, Exception('Failed to add DN! LDAP error! Reason: %s Diag: %s' % (
+						return False, LDAPAddException(
+							entry,
 							message['protocolOp']['resultCode'],
-							message['protocolOp']['diagnosticMessage'])
+							message['protocolOp']['diagnosticMessage']
 						)
 
 			return True, None
 		except Exception as e:
 			return False, e
 
-	async def modify(self, entry, changes, controls = None):
+	async def modify(self, entry:str, changes:Dict[str, object], controls:List[Control] = None):
 		"""
 		Performs the modify operation.
 		
@@ -491,16 +586,17 @@ class MSLDAPClientConnection:
 				message = message.native
 				if msg_type == 'modifyResponse':
 					if message['protocolOp']['resultCode'] != 'success':
-						return False, Exception('Failed to add DN! LDAP error! Reason: %s Diag: %s' % (
+						return False, LDAPModifyException(
+							entry,
 							message['protocolOp']['resultCode'],
-							message['protocolOp']['diagnosticMessage'])
+							message['protocolOp']['diagnosticMessage']
 						)
 
 			return True, None
 		except Exception as e:
 			return False, e
 
-	async def delete(self, entry):
+	async def delete(self, entry:str):
 		"""
 		Performs the delete operation.
 		
@@ -523,16 +619,17 @@ class MSLDAPClientConnection:
 				message = message.native
 				if msg_type == 'delResponse':
 					if message['protocolOp']['resultCode'] != 'success':
-						return False, Exception('Failed to add DN! LDAP error! Reason: %s Diag: %s' % (
+						return False, LDAPDeleteException(
+							entry,
 							message['protocolOp']['resultCode'],
-							message['protocolOp']['diagnosticMessage'])
+							message['protocolOp']['diagnosticMessage']
 						)
 
 			return True, None
 		except Exception as e:
 			return False, e
 	
-	async def search(self, base, query, attributes, search_scope = 2, size_limit = 1000, types_only = False, derefAliases = 0, timeLimit = None, controls = None, return_done = False):
+	async def search(self, base:str, query:str, attributes:List[str], search_scope:int = 2, size_limit:int = 1000, types_only:bool = False, derefAliases:int = 0, timeLimit:int = None, controls:List[Control] = None, return_done:bool = False):
 		"""
 		Performs the search operation.
 		
@@ -560,7 +657,7 @@ class MSLDAPClientConnection:
 		:return: Async generator which yields (`LDAPMessage`, None) tuple on success or (None, `Exception`) on error
 		:rtype: Iterator[(:class:`LDAPMessage`, :class:`Exception`)]
 		"""
-		if self.status != MSLDAPClientStatus.RUNNING:
+		if self.status not in [MSLDAPClientStatus.CONNECTED, MSLDAPClientStatus.RUNNING]:
 			yield None, Exception('Connection not running! Probably encountered an error')
 			return
 		try:
@@ -613,7 +710,7 @@ class MSLDAPClientConnection:
 		except Exception as e:
 			yield (None, e)
 
-	async def pagedsearch(self, base, query, attributes, search_scope = 2, size_limit = 1000, typesOnly = False, derefAliases = 0, timeLimit = None, controls = None):
+	async def pagedsearch(self, base:str, query:str, attributes:List[str], search_scope:int = 2, size_limit:int = 1000, typesOnly:bool = False, derefAliases:bool = 0, timeLimit:int = None, controls:List[Control] = None, rate_limit:int = 0):
 		"""
 		Paged search is the same as the search operation and uses it under the hood. Adds automatic control to read all results in a paged manner.
 		
@@ -635,6 +732,8 @@ class MSLDAPClientConnection:
 		:type timeLimit: int
 		:param controls: additional controls to be passed in the query
 		:type controls: dict
+		:param rate_limit: time to sleep bwetween each query
+		:type rate_limit: float
 		:return: Async generator which yields (`dict`, None) tuple on success or (None, `Exception`) on error
 		:rtype: Iterator[(:class:`dict`, :class:`Exception`)]
 		"""
@@ -645,7 +744,7 @@ class MSLDAPClientConnection:
 		try:
 			cookie = b''
 			while True:
-				
+				await asyncio.sleep(rate_limit)
 				ctrl_list_temp = [
 					Control({
 						'controlType' : b'1.2.840.113556.1.4.319',
@@ -680,6 +779,8 @@ class MSLDAPClientConnection:
 							return
 						
 						if 'resultCode' in res['protocolOp']:
+							if res['protocolOp']['resultCode'] != 'success':
+								raise LDAPSearchException(res['protocolOp']['resultCode'], res['protocolOp']['diagnosticMessage'])
 							for control in res['controls']:
 								if control['controlType'] == b'1.2.840.113556.1.4.319':
 									try:
@@ -698,9 +799,32 @@ class MSLDAPClientConnection:
 		except Exception as e:
 			yield (None, e)
 
+	async def whoami(self):
+		if self.status != MSLDAPClientStatus.RUNNING:
+			return None, Exception('Connection not running! Probably encountered an error')
+		
+		ext = {
+			'requestName': b'1.3.6.1.4.1.4203.1.11.3',
+		}
+		br = { 'extendedReq' : ExtendedRequest(ext)}
+		msg = { 'protocolOp' : protocolOp(br)}
+
+		msg_id = await self.send_message(msg)
+		res = await self.recv_message(msg_id)
+		res = res[0]
+		if isinstance(res, Exception):
+			return None, res
+		if res.native['protocolOp']['resultCode'] != 'success':
+			return False, LDAPBindException(
+					res['protocolOp']['resultCode'], 
+					res['protocolOp']['diagnosticMessage']
+				)
+		return res.native['protocolOp']['responseValue'].decode(), None
+		
+
 
 	async def get_serverinfo(self):
-		if self.status != MSLDAPClientStatus.RUNNING:
+		if self.status not in [MSLDAPClientStatus.CONNECTED, MSLDAPClientStatus.RUNNING]:
 			return None, Exception('Connection not running! Probably encountered an error')
 
 		attributes = [
@@ -742,85 +866,4 @@ class MSLDAPClientConnection:
 		if isinstance(res, Exception):
 			return None, res
 		
-		#print('res')
-		#print(res)
 		return convert_attributes(res.native['protocolOp']['attributes']), None
-
-
-async def amain():
-	import traceback
-	from msldap.commons.url import MSLDAPURLDecoder
-
-	base = 'DC=TEST,DC=CORP'
-
-	#ip = 'WIN2019AD'
-	#domain = 'TEST'
-	#username = 'victim'
-	#password = 'Passw0rd!1'
-	##auth_method = LDAPAuthProtocol.SICILY
-	#auth_method = LDAPAuthProtocol.SIMPLE
-
-	#cred = MSLDAPCredential(domain, username, password , auth_method)
-	#target = MSLDAPTarget(ip)
-	#target.dc_ip = '10.10.10.2'
-	#target.domain = 'TEST'
-
-	url = 'ldaps+ntlm-password://test\\Administrator:QLFbT8zkiFGlJuf0B3Qq@WIN2019AD/?dc=10.10.10.2'
-
-	dec = MSLDAPURLDecoder(url)
-	cred = dec.get_credential()
-	target = dec.get_target()
-
-	print(cred)
-	print(target)
-
-	input()
-
-	client = MSLDAPClientConnection(target, cred)
-	await client.connect()
-	res, err = await client.bind()
-	if err is not None:
-		raise err
-	
-	user = "CN=ldaptest_2,CN=Users,DC=test,DC=corp"
-	#attributes = {'objectClass':  ['inetOrgPerson', 'posixGroup', 'top'], 'sn': 'user_sn', 'gidNumber': 0}
-	#res, err = await client.add(user, attributes)
-	#if err is not None:
-	#	print(err)
-
-	#changes = {
-	#	'unicodePwd': [('replace', ['"TESTPassw0rd!1"'])],
-	#	#'lockoutTime': [('replace', [0])]
-	#}
-
-	#res, err = await client.modify(user, changes)
-	#if err is not None:
-	#	print('ERR! %s' % err)
-	#else:
-	#	print('OK!')
-	
-	res, err = await client.delete(user)
-	if err is not None:
-		print('ERR! %s' % err)
-	
-	await client.disconnect()
-
-
-
-if __name__ == '__main__':
-	from msldap import logger
-	from msldap.commons.credential import MSLDAPCredential, LDAPAuthProtocol
-	from msldap.commons.target import MSLDAPTarget
-	from msldap.protocol.query import query_syntax_converter
-
-	logger.setLevel(2)
-
-
-	asyncio.run(amain())
-
-	
-
-			
-			
-
-		
\ No newline at end of file
